/*mmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmm
			SVGA256 GRAPHICS LIBRARY
	     Copyright (C) 1993 by Orthware.   All rights reserved.
/*mmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmm*/


/**************************************************************************
/*	DEMOMSGS.C	Misc routines to display pop-up messages and do
/*			user input of strings and numbers.
/*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <math.h>
#include "svga256.h"
#include "s256demo.h"

#define BYTE			unsigned char
#define WORD			unsigned short 
#define LONG			unsigned long

#define NULLCHAR		'\0'

#define	TICKS			(*(LONG far *) 0x0040006c)  /* system timer */
#define FALSE			0
#define TRUE			1
#define	ESC			27
#define	BS			8
#define	ENTER			13
#define	DEL			211

int getkeypress(void);

/************************************************************************
* global variables (local to this module)
/***********************************************************************/

/* vars holding character coords of a "centered" message box */
/* once a center message is opened you may want to reference */
/* these for further screen updating relative to the window. */

int  cmsg_lt, cmsg_top, cmsg_rt, cmsg_bot;


/***********************************************************************
 * Message strings for Graphics mode. Each set of strings must end with a
 * NULL string which acts as a delimiter between each message for routines.
/***********************************************************************/
char *msg_strings[] =
{
/*  0*/	"       SVGA256 DEMO OPTIONS",
	" ",
	"  0) TEXT demo.",
	"  1) RECTANGLE demo.",
	"  2) PALETTE demo.",
	"  3) PIXEL demo.",
	"  4) \"PUT\" LINE demo.",
	"  5) \"XOR\" LINE demo.",
	"  6) CIRCLE/ELLIPSE demo.         ",
	"  7) WINDOWING demo.",
	"  8) IMAGE display demo.",
	"  9) EXIT to DOS.",
	"",
/*  1*/ " Press key for main menu. ",
	"",
/*  2*/ " Error reading image file, press any key. ",
	"",
/*  3*/ " Memory allocation error, press any key. ",
	"",
/*  4*/ " Reading image ... ",
	"",
/*  5*/ "Demo Image is too big for this mode!",
	"",
};


/***************************************************************************
/*  centermsg:		Opens a window in the center of the screen containing
/*			the "msgno" string set.  If "waitforkey" is nonzero
/*			it waits for a key press and returns the key code,
/*			else if it is zero the routine does not wait for a 
/*			keypress and it returns the window handle so you 
/*			can close the window later.
/***************************************************************************/
int  centermsg(int  msgno, int waitforkey)
{
	int  chc, chr, wnum, strlen, strct, key;
	int  framecolor, fillcolor, textcolor;


	/* set up colors */

	framecolor = BORDER_CLR;
	fillcolor = MSGFILL_CLR;
	textcolor = MSGTEXT_CLR;


	/* get number of strings and maximum string length of the message */

	getmsgsize(msgno, &strct, &strlen);


	/* set up coordinates and open the window */

	cmsg_lt = chc = SV_charmaxcol() / 2 - strlen / 2 - 1;
	cmsg_top = chr = SV_charmaxrow() / 2 - strct / 2 - 1;

	wnum = SV_openwind(TRUE, chc, chr, (cmsg_rt = chc + strlen + 1),
		(cmsg_bot = chr + strct + 1), framecolor, fillcolor,
		SHADOW_CLR);


	/* display the message in the window */

	if (wnum > 0)
		dispmsg(msgno, chc + 1, chr + 1, textcolor);


	/* either wait for key press, or do not and return window number */

	if (waitforkey)
	{
		key = 0;
		while(!key) key = getkeypress();
		SV_closewind(wnum);
		return key;
	}
	else
		return wnum;
}


/******************************************************************************
/* msg_stringindex	Given a "msgno" this routine returns the index
/*			of the 1st string in the message associated with msgno.
/******************************************************************************/
int  msg_stringindex(int msgno)
{
	int  i, j;

	for (i = 0, j = 0; i < msgno; i++)
		while(*msg_strings[j++]);

	return j;
}

/***************************************************************************
/*   dispmsg:		Takes a message number (index into the array of
/*			string sets "msg_strings") and displays them
/*			starting at char coordinates (NOT pixel coords) 
/*			"chc", "chr"  (char column, char row).
/***************************************************************************/
void  dispmsg(int msgno, int chc, int chr, int color)
{
	int  i, col, row;


	i = msg_stringindex(msgno);

	while(*msg_strings[i])
	{
		col = chc;  row = chr++;
		dispstring(msg_strings[i++], &col, &row, color);
	}
}


/***************************************************************************
/*   getmsgsize:	Given an index of a set of strings in "msg_strings"
/*			this routine returns the number of strings in "stringct"
/*			and the length (# of chars) of the longest string in
/*			"msglength".
/**************************************************************************/
void  getmsgsize(int msgno, int *stringct, int *msglength)
{
	int  i, templen;


	/* GET INDEX OF 1ST STRING IN MSG */

	i = msg_stringindex(msgno);


	/* GET 1) NUM OF STRINGS, 2) MAX STRING LENGTH */
	*stringct = 0;
	*msglength = 0;

	while(*msg_strings[i])
	{
		(*stringct)++;
		templen = strlen(msg_strings[i++]);
		if (templen > *msglength)  *msglength = templen;
	}
}

/***************************************************************************
/* enterstring:   - Allows input of a character string on the graphics screen
/* 		  - Accepts input of a string from user followed by a <CR>.
/*		  - Returns users last key press code.
/*		  - "maxlen" has a maximum of 80 chars.
/*		  - String is unchanged if ESC is pressed or if CR is pressed
/*		    when no other input has occurred.
/*		  - The "acceptcodes" are key codes that act the same as
/*		    the ENTER key (UP ARROW & DN ARROW might be these values
/*		    to exit this routine and move to a next field)
/***************************************************************************/
int  enterstring(char *str, int chc, int chr, int maxlen, int acceptcode1,
	int acceptcode2, int acceptcode3, int acceptcode4)
{
	BYTE  tempstr[81];
	int  col, beforecolor, inputcolor, savedblock;
	int  mx, my, lastx, lasty, i=0, key=0, firstkey=1, onscreen=1;
	long ticks;

/* macro to display the cursor */
#define	SCURSOR(clr) 	(inputcursor(col - (i >= maxlen), chr, clr))


	tempstr[0] = NULLCHAR;

	col = chc;
	SV_charmoveto(col, chr);
	beforecolor = SV_readpixel(SV_getx(), SV_gety());
	inputcolor = SV_BF(INPUTBKG_CLR, INPUTFORE_CLR);  /* set bkg/fgd color*/


	/* DRAW A REVERSE VIDEO INPUT AREA FOR STRING WITH ORIGINAL*/
	/* SUPPLIED STRING DISPLAYED IN THIS AREA, THEN ENTER INPUT LOOP */
	/* THE SCREEN AREA UNDER THE STRING IS SAVED BY SV_saveblock().  */

	lasty = SV_chary(chr) + SV_charht();
	lastx = SV_charx(col) + maxlen * SV_charwidth() - 1;
	savedblock = SV_saveblock(SV_charx(col), lasty, lastx, lasty);
	SV_line(SV_charx(col), lasty, lastx, lasty, INPUTBKG_CLR);

	SV_clearstring(maxlen, INPUTBKG_CLR);
	dispstring(str, &col, &chr, inputcolor);
	col = chc;
	SV_charmoveto(col, chr);
	SCURSOR(CURSOR_CLR);


	/* INPUT CHARACTERS TILL ESC OR AN "ACCEPT" KEY IS PRESSED */

	while(1)
	{
		if (firstkey)
		{
			if (*str)	/* wait only if string is nonempty */
			{
				/* wait for the first key stroke */
				while(!key)
					key = getkeypress();

				/* clear the field for new input */
				SV_clearstring(maxlen, INPUTBKG_CLR);
				SCURSOR(CURSOR_CLR);
			}

			ticks = TICKS;
			firstkey = 0;
		}
		else
			key = getkeypress();   /* get the next key */


		/* check for an exit condition */
		if (key)
		  if ((key == ENTER) || (key == acceptcode1) ||
		     (key == acceptcode2) || (key == acceptcode3) ||
		     (key == acceptcode4) || (key == ESC))
				break;


		/* check if cursor blinking needs to be updated */
		if (TICKS - ticks > 2)
		{
			SCURSOR((onscreen ^= 1) ? CURSOR_CLR : INPUTBKG_CLR);
			ticks = TICKS;
		}


		/* process the keystroke */

		if ((key >= 32) && (key <= 126))
		{
			SCURSOR(INPUTBKG_CLR);
			if (i >= maxlen)
			{
				 --i;
				 --col;
			}
			tempstr[i] = (BYTE) key;
			dispchar(key, &col, &chr, inputcolor);
			i++;
			SCURSOR(CURSOR_CLR);
		}
		else if ((key == BS) && (i > 0))
		{
			SCURSOR(INPUTBKG_CLR);
			col--; i--;
			dispchar(' ', &col, &chr, inputcolor);
			col--;
			SCURSOR(CURSOR_CLR);
		}
		else if (key == DEL)
		{
			SCURSOR(INPUTBKG_CLR);
			col = chc;
			SV_charmoveto(col, chr);
			SV_clearstring(i, inputcolor);
			i = 0;
			SCURSOR(CURSOR_CLR);

			tempstr[0] = NULLCHAR;
		}

	}

	tempstr[i] = NULLCHAR;

	if ((key != ESC) && tempstr[0] && (i <= maxlen))
		strcpy(str, tempstr);

	SV_charmoveto(chc, chr);
	SV_clearstring(maxlen, beforecolor);
	SV_closeblock(savedblock);
	dispstring(str, &chc, &chr, INFO_CLR);

	return key;
}

/****************************************************************************
/* enterfloat:  This procedure is used to enter a dbl-precision floating point
/* "value" from the keyboard onto graphics screen.  The value is to have a max
/* of "wlen" digits to the left of the decimal point, and "flen" digits to the
/* right.  The value must be in the range between (and including) "fmin"
/* through "fmax."  The value will be entered at the specified "chr"
/* and "chc".  Procedure returns the keycode of the last key press.
/*  - The "acceptcodes" are key codes that act the same as
/*    the ENTER key (UP ARROW & DN ARROW might be these values
/*    to exit this routine and move to a next field)
/****************************************************************************/
int  enterfloat(double *value, int wlen, int flen, double fmin,
	double fmax, int chc, int chr, int acceptcode1, int acceptcode2,
	int acceptcode3, int acceptcode4)
{
	char  string[20];
	int   i, field, wdigs, fdigs, row, col, key=0, firstkey=1, onscreen=1;
	int   beforecolor, inputcolor, savedblock;
	int   mx, my, lastx, lasty;
	double wval, fval, val;
	long  ticks;

/* macro to display next digit and advance "col" index */
#define	DISPDIG	{dispchar(key, &col, &row, inputcolor);}

/* macro to display the cursor */
#define FCURSOR(clr)	(inputcursor(col - ((col - chc) >= field), chr, clr))



	row = chr; col = chc;
	SV_charmoveto(col, row);
	beforecolor = SV_readpixel(SV_getx(), SV_gety());
	inputcolor = SV_BF(INPUTBKG_CLR, INPUTFORE_CLR);

	/* initialize the digit counters */
	wdigs = 0;
	fdigs = 0;
	field = wlen + flen + (flen > 0);

	/* DRAW A REVERSE VIDEO INPUT AREA FOR STRING WITH ORIGINAL*/
	/* SUPPLIED VALUE DISPLAYED IN THIS AREA, THEN ENTER INPUT LOOP*/

	lasty = SV_chary(row) + SV_charht();
	lastx = SV_charx(col) + field * SV_charwidth() - 1;

	savedblock = SV_saveblock(SV_charx(col), lasty, lastx, lasty);
	SV_line(SV_charx(col), lasty, lastx, lasty, INPUTBKG_CLR);

	SV_clearstring(field, INPUTBKG_CLR);
	dispstring(floatstring(string, *value, wlen , flen, FALSE),
		&col, &row, inputcolor);
	string[0] = NULLCHAR;
	row = chr; col = chc;
	SV_charmoveto(col, row);
	FCURSOR(CURSOR_CLR);


	/* INPUT CHARACTERS TILL ESC OR AN "ACCEPT" KEY IS PRESSED */

	while(1)
	{
		if (firstkey)
		{
			/* wait for the first key stroke */
			while(!key)
				key = getkeypress();

			/* clear the field for new input */
			SV_clearstring(field, INPUTBKG_CLR);
			FCURSOR(CURSOR_CLR);

			ticks = TICKS;
			firstkey = 0;
		}
		else
			key = getkeypress();   /* get the next key */


		/* check for an exit condition */
		if (key)
		  if ((key == ENTER) || (key == acceptcode1) ||
		     (key == acceptcode2) || (key == acceptcode3) ||
		     (key == acceptcode4) || (key == ESC))
				break;


		/* check if cursor blinking needs to be updated */
		if (TICKS - ticks > 2)
		{
			FCURSOR((onscreen) ? CURSOR_CLR : INPUTBKG_CLR);
			onscreen = 1 - onscreen;
			ticks = TICKS;
		}


		/* process the keystroke */
		if ((key >= '0') && (key <= '9'))
		{
			if (wdigs + fdigs >= field)
			{
				if (fdigs) --fdigs;
				else  -- wdigs;

				col--;
			}

			if (fdigs && (fdigs < flen + 1))
			{
				FCURSOR(INPUTBKG_CLR);
				DISPDIG
				string[wdigs + fdigs++] = (BYTE) key;
				FCURSOR(CURSOR_CLR);
                        }

                        else if ((fdigs == 0) && (wdigs < wlen))
			{
				FCURSOR(INPUTBKG_CLR);
				DISPDIG
				string[wdigs++] = (BYTE) key;
				FCURSOR(CURSOR_CLR);
                        }
		}

		else if ((key == '-') && (wlen > 1) && (wdigs + fdigs == 0))
		{
			FCURSOR(INPUTBKG_CLR);
			DISPDIG
			string[wdigs++] = (BYTE)key;
			FCURSOR(CURSOR_CLR);
		}

		else if ((key == '.') && (flen > 0) && (fdigs == 0))
		{
			FCURSOR(INPUTBKG_CLR);
			DISPDIG
			string[wdigs + fdigs++] = (BYTE) key;
			FCURSOR(CURSOR_CLR);
		}

		else if ((key == BS) && (wdigs + fdigs > 0))
		{
			FCURSOR(INPUTBKG_CLR);
			--col;
			dispchar(' ', &col, &row, inputcolor);
			--col;
			(fdigs) ? --fdigs : --wdigs;
			FCURSOR(CURSOR_CLR);
		}

		else if (key == DEL)
		{
			FCURSOR(INPUTBKG_CLR);
			wdigs = 0;
			fdigs = 0;

			row = chr; col = chc;
			SV_charmoveto(col, row);
			SV_clearstring(field, INPUTBKG_CLR);
			FCURSOR(CURSOR_CLR);
		}

	}

	/* convert entered string into a double-prec floating point value */

	if ((wdigs + fdigs > 0) && (key != ESC))
        {
                wval = 0.0;     /* initialize the 'whole' part */

                for (i = (string[0] == '-'); i < wdigs; ++i)
                        wval = wval * 10.0 + (string[i] - '0');

                fval = 0.0;     /* inititalize the 'fractional' part */

                for (i = wdigs + fdigs; i > wdigs + 1; --i)
                        fval = fval / 10.0 + (string[i - 1] - '0');

                val = wval + fval / 10.0;

                if (string[0] == '-') val = -val;

		if ((val >= fmin) && (val <= fmax))
			*value = val;
	}

	SV_charmoveto(chc, chr);
	SV_clearstring(field, beforecolor);
	SV_closeblock(savedblock);
	dispstring(floatstring(string, *value, wlen , flen, FALSE),
		&chc, &chr, INFO_CLR);

	return key;
}

/****************************************************************************
/*  enterdec:   for inputting a decimal in graphics mode.  "value" is
/*		allocated storage for the integer decimal value. "length"
/*		the maximum length of characters to enter. "dmin" and "dmax"
/*		is the inclusive range of the value to be entered.
/*		"chc" and "chr" are the character coordinates of
/*		the input area.
/****************************************************************************/
int  enterdec(int *value, int length, int dmin, int dmax, int chc, int chr,
	int acceptcode1, int acceptcode2, int acceptcode3, int acceptcode4)
{
	int  key;
	double  fvalue;

	fvalue = (double) *value;      /* convert 'int' value to 'float' */

	key = enterfloat(&fvalue, length, 0, (double) dmin, (double) dmax,
		chc, chr, acceptcode1, acceptcode2, acceptcode3, acceptcode4);

	*value = (int) fvalue;        /* convert 'float' back to 'int' */

	return key;
}

/*************************************************************************
/* inputcursor:		Draws a cursor for character input positions.
/*************************************************************************/
void  inputcursor(int chc, int chr, int color)
{
	register int  x0, x1, y;

	x0 = SV_charx(chc);
	x1 = x0 + SV_charwidth() - 1; 
	y = SV_chary(chr) + SV_charht() - 1;
	SV_scanline(x0, x1, y++, color);
	SV_scanline(x0, x1, y, color);
}

/***************************************************************************
/* floatstring:  This procedure is used to convert the double-precision
/* floating "value" specified into a "string" of characters in the specified
/* format. "wlength" is number of digits to left of dec (including the "-")
/* and "flength" is number of digits to right of the decimal point.
/* The "format" flag has bit-set meanings:
/*		bit 0 = leading zeros displayed.
/****************************************************************************/
char *floatstring(char *string, double value, int wlength,
	int flength, int format)
{
	double  work_value, fvalue, roundup;
	long  wvalue;
	register short int  i, non_zero, offset;
	register BYTE  work_digit;


	roundup = 0.5;

	for (i = 0; i < flength; ++i) roundup = roundup / 10.0;

	offset = -1;

	work_value = ((value < 0.0) ? -value : value) + roundup;

	wvalue = (long) work_value;
	fvalue = 10.0 * (work_value - wvalue);

	non_zero = wlength + offset + (wlength == 0);

	for (i = wlength; i > 0; --i)
	{
		work_digit = wvalue % 10;
		if (work_digit) non_zero = i + offset;
		string[i + offset] = work_digit + '0';
		wvalue = wvalue / 10;
	}

	if (flength > 0)
	{
		string[wlength + offset + 1] = '.';
		for (i = 0; i < flength; ++i)
		{
			work_digit = fvalue;
			string[i + wlength + offset + 2] = work_digit + '0';
			fvalue = 10.0 * (fvalue - work_digit);
		}
	}

	if (value < 0.0) string[--non_zero] = '-';

	if (!(format & 1))
		 for (i = 0; i < non_zero; ++i) string[i] = ' ';

	string[wlength + flength + (flength > 0) + offset + 1] = NULLCHAR;

	return string;
}

/**************************************************************************
/*   dispchar:		Displays "char" at the CHARACTER posn "*chc","*chr"
/*			with the color "color" where the lower 8 bits of
/*			color is a foreground color and the next 8 bits are
/*			a background color.  Background pixels are not drawn
/*			if background color is 0. This routine is VGA & 80386
/*			specific.  "*chc" is increased by 1. 
/*			The current graphics pointer coordinates are also
/*			updated to the position at end of the char.
/*************************************************************************/
void  dispchar(int  ch, int *chc, int *chr, int color)
{
	SV_putchar(ch, SV_charx(*chc), SV_chary(*chr), color);

	SV_movetoxy(SV_charx(*chc) + SV_charwidth(), SV_chary(*chr));

	*chc += 1;
}

/**************************************************************************
/*   dispstring:	Displays "string" at the CHARACTER posn "chc","chr"
/*			with the color "color" where the lower 8 bits of
/*			color is a foreground color and the next 8 bits are
/*			a background color.  Background pixels are not drawn
/*			if background color is 0. This routine is VGA & 80386
/*			specific.  "*chc" is increased by the length of string. 
/*			The current graphics pointer coordinates are also
/*			updated to the end of the string.
/*************************************************************************/
void  dispstring(char *string, int *chc, int *chr, int color)
{
	SV_putstringxy(string, SV_charx(*chc), SV_chary(*chr), color);

	SV_movetoxy(SV_charx(*chc) + SV_stringwidth(string), SV_chary(*chr));

	*chc += strlen(string);
}


